package org.netkernel.mod.architecture.bits;

import java.io.File;
import java.text.DecimalFormat;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;

import org.netkernel.container.ISpaceListener;
import org.netkernel.ext.system.representation.IRepDeployedModules;
import org.netkernel.layer0.boot.BootUtils;
import org.netkernel.layer0.meta.IEndpointMeta;
import org.netkernel.layer0.meta.IModuleMeta;
import org.netkernel.layer0.meta.IPhysicalEndpointMeta;
import org.netkernel.layer0.meta.IPrototypeMeta;
import org.netkernel.layer0.meta.IRepresentationMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.nkf.INKFRequest;
import org.netkernel.layer0.nkf.INKFRequestContext;
import org.netkernel.layer0.nkf.INKFResponse;
import org.netkernel.layer0.nkf.INKFResponseReadOnly;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IHDSNodeList;
import org.netkernel.layer0.representation.impl.HDSBuilder;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.IMetaRepresentation;

public class SearchAccessor extends StandardAccessorImpl implements ISpaceListener
{
	private String mIndexDir;
	private HDSBuilder mIndexSpecificationBuilder;
	private int mIndexCount;
	private AtomicBoolean mIndexBuilt=new AtomicBoolean();
	
	public SearchAccessor()
	{	this.declareThreadSafe();
	}
	
	public void postCommission(INKFRequestContext aContext) throws Exception
	{	aContext.getKernelContext().getKernel().addSpaceListener(this);
	}
	public void preDecommission(INKFRequestContext aContext) throws Exception
	{	aContext.getKernelContext().getKernel().removeSpaceListener(this);
	}
	
	public void onSource(INKFRequestContext aContext) throws Exception
	{
		String type=aContext.getThisRequest().getArgumentValue("type");
		if (type.equals("Index"))
		{	onSearchIndex(aContext);
		}
		else if (type.equals("RawResults"))
		{	onRawResults(aContext);
		}
	}
	
	public void onRawResults(INKFRequestContext aContext) throws Exception
	{
		lazyInitialiseIndex(aContext,true);
		
		String entityId=aContext.getThisRequest().getArgumentValue("entityId");
		
		
		INKFRequest req=aContext.createRequest("active:textSearch");
		req.addArgument("directory", mIndexDir);
		req.addArgument("limit", "50");
		req.addArgumentByValue("operand", "d1:("+entityId+") d2:("+entityId+") d1:("+entityId+"*)" );
		IHDSNode rs=(IHDSNode)aContext.issueRequest(req);
		
		//boost different entities differently
		IHDSNodeList hits=rs.getNodes("/hits/hit");
		Map<String,Float> scores = new HashMap<String,Float>();
		float maxScore=0.0f;
		DecimalFormat df = new DecimalFormat("0.0000");
		for (IHDSNode hit: hits)
		{
			float score=df.parse((String)hit.getFirstValue("score")).floatValue();
			String hash=(String)hit.getFirstValue("field[1]/value");
			String type=(String)hit.getFirstValue("field[2]/value");
			float weight=1.0f;
			if (type.equals("M")) weight=4.0f;
			if (type.equals("P")) weight=4.0f;
			if (type.equals("R")) weight=3.0f;
			if (type.equals("S")) weight=1.5f;
			if (type.equals("E")) weight=0.5f;
			score*=weight;
			if (score>maxScore) maxScore=score;
			scores.put(hash,score);
		}
		
		IRepDeployedModules modules=aContext.source("active:moduleList", IRepDeployedModules.class);
		
		HDSBuilder b=new HDSBuilder();
		b.pushNode("hits");
		for (IHDSNode hit: hits)
		{
			String hash=(String)hit.getFirstValue("field[1]/value");
			float score=0.02f*scores.get(hash); //maxScore;
			
			if (score>0.002f)
			{
				b.pushNode("hit");
				b.addNode("score",Float.toString(score));
				String type=(String)hit.getFirstValue("field[2]/value");
				
				IMetaRepresentation meta=modules.metadataForHash(hash);
				if (meta!=null)
				{	addResultField(b,"resolverID","explorerSearchResolver-"+type);
					addResultField(b,"documentID",hash);
					addResultField(b,"title",meta.getName());
					addResultField(b,"desc",meta.getDescription());
				}
				b.popNode();
			}
		}
		aContext.createResponseFrom(b.getRoot());
	}
	
	private void addResultField(HDSBuilder b, String aName, String aValue)
	{	b.pushNode("field");
		b.addNode("name",aName);
		b.addNode("value",aValue);
		b.popNode();
	}
	
	private void lazyInitialiseIndex(INKFRequestContext aContext, boolean aWait) throws Exception
	{	if (mIndexDir==null)
		{	
			INKFRequest req=aContext.createRequest("res:/aebits/view/searchIndex");
			if (aWait)
			{	aContext.issueRequest(req);
			}
			else
			{	aContext.issueAsyncRequest(req);
			}
		}
		else if (!mIndexBuilt.get() && aWait)
		{	while (!mIndexBuilt.get())
			{	Thread.sleep(50);
			}
		}
	}
	
	public void onSearchIndex(INKFRequestContext aContext) throws Exception
	{
		String indexDirName="ExplorerSearchIndex"+System.currentTimeMillis();
		File temp=File.createTempFile(indexDirName, null);
		File tempDir=temp.getParentFile();
		temp.delete();
		File searchDirFile=new File(tempDir, indexDirName+File.pathSeparator);
		searchDirFile.mkdir();
		String searchDir=searchDirFile.toURI().toString();
		mIndexDir=searchDir;
		resetIndex();
		
		IRepDeployedModules modules=aContext.source("active:moduleList", IRepDeployedModules.class);
		IHDSNode list=modules.getHierarchy();
		for (IHDSNode module: list.getNodes("/modules/module"))
		{	String mHash=(String)module.getFirstValue("hash");
			IModuleMeta moduleMeta=(IModuleMeta)modules.metadataForHash(mHash);
			indexItem(moduleMeta,null,mHash,aContext);
			for (IHDSNode space: module.getNodes("spaces//space"))
			{	String spaceHash=(String)space.getFirstValue("hash");
				ISpaceMeta spaceMeta=(ISpaceMeta)modules.metadataForHash(spaceHash);
				indexItem(spaceMeta,null,spaceHash, aContext);
				ISpaceElements se=spaceMeta.getElements();
				IEndpoint lastEp=null;
				for (int i=0; i<se.size(); i++)
				{	IEndpoint ep=se.getPhysicalEndpoint(i);
					IIdentifier eid=se.getIdentifier(i);
					if (ep!=lastEp)
					{	lastEp=ep;
						String epHash=modules.hashForComponent(ep);
						IMetaRepresentation epMeta=modules.metadataForHash(epHash);
						indexItem(epMeta,eid,epHash,aContext);
					}
					
					if (se.getDelegatedSpace(i)==null)
					{	if (eid!=null)
						{	String eHash=spaceHash+"/"+eid.toString();
							IMetaRepresentation eMeta=modules.metadataForHash(eHash);
							indexItem(eMeta,eid,eHash,aContext);
						}
					}
				}
			}
		}
		writeIndex(aContext);
		mIndexBuilt.set(true);
	}
	
	private void resetIndex()
	{	mIndexCount=0;
		mIndexSpecificationBuilder=new HDSBuilder();
		mIndexSpecificationBuilder.pushNode("batch");
	}
	
	private String indexItem(IMetaRepresentation aMeta, IIdentifier aId, String aHash, INKFRequestContext aContext) throws Exception
	{
		StringBuilder sb1=new StringBuilder(128);
		StringBuilder sb2=new StringBuilder(128);

		String rawName=aMeta.getName();
		sb1.append(rawName);
		sb1.append(' ');
		String camelName=unCamelize(rawName);
		sb2.append(camelName);
		sb2.append(' ');
		
		if (aId!=null)
		{	String id=aId.toString().replaceAll("\\:", " ");
			sb2.append(id);
			sb2.append(' ');
		}
		
		if (aMeta.getDescription()!=null)
		{	sb1.append(aMeta.getDescription());
			sb1.append(' ');
		}
		String type="";
		if (aMeta instanceof IModuleMeta)
		{	sb1.append(((IModuleMeta)aMeta).getSource().toString());
			type="M";
		}
		else if (aMeta instanceof IPhysicalEndpointMeta)
		{	type="E";
		}
		else if (aMeta instanceof ISpaceMeta)
		{	type="S";
		}
		else if (aMeta instanceof IRepresentationMeta)
		{	
			String className=((IRepresentationMeta)aMeta).getImplementation().getName();
			className=className.replaceAll("\\.", " ");
			sb1.append(className);
			type="R";
		}
		else if (aMeta instanceof IPrototypeMeta)
		{	String className=((IPrototypeMeta)aMeta).getImplementation().getName();
			className=className.replaceAll("\\.", " ");
			sb1.append(className);
			sb1.append(' ');
			sb1.append(aHash);
			type="P";
		}
		else if (aMeta instanceof IEndpointMeta)
		{
			String rawGrammar=((IEndpointMeta)aMeta).getIdentifierGrammar().toString();
			sb1.append(rawGrammar);
			sb1.append(' ');
			String grammar=unCamelize(rawGrammar);
			//if (!grammar.equals(rawGrammar))
			{	sb2.append(grammar);
			}
			sb1.append(' ');
			sb1.append(aHash);
			type="L";
		}
		mIndexSpecificationBuilder.pushNode("item");
		addIndexField("hash",aHash,"NO","YES");
		addIndexField("type",type,"NO","YES");
		addIndexField("d1",sb1.toString(),"ANALYZED","NO");
		addIndexField("d2",sb2.toString(),"ANALYZED","NO");
		mIndexSpecificationBuilder.popNode();
		mIndexCount++;
		
		if (mIndexCount>399)
		{
			writeIndex(aContext);
			resetIndex();
		}
		
		
		return sb1.toString();
	}
	
	private void addIndexField(String aName, String aValue, String aAnalyze, String aStore)
	{
		mIndexSpecificationBuilder.pushNode("field");
		mIndexSpecificationBuilder.addNode("@index", aAnalyze);
		mIndexSpecificationBuilder.addNode("@store", aStore);
		mIndexSpecificationBuilder.addNode("name", aName);
		mIndexSpecificationBuilder.addNode("value", aValue);
		mIndexSpecificationBuilder.popNode();
	}
	
	private void writeIndex(INKFRequestContext aContext) throws Exception
	{	INKFRequest req=aContext.createRequest("active:textIndexBatch");
		req.addArgument("directory", mIndexDir);
		req.addArgumentByValue("operand", mIndexSpecificationBuilder.getRoot());
		aContext.issueRequest(req);
	}
	
	private static String unCamelize(String aInput)
	{
		//transition from lower->upper or char->number or number->char causes split
		StringBuilder result=new StringBuilder();
		boolean lastUpper=false;
		boolean lastDigit=false;
		for (int i=0; i<aInput.length(); i++)
		{
			char c=aInput.charAt(i);
			boolean upper=Character.isUpperCase(c);
			boolean digit=Character.isDigit(c);
			if (i>0)
			{
				if ( (upper && !lastUpper) || (digit!=lastDigit) )
				{	result.append(" ");
				}
				if (lastUpper && upper && i<aInput.length()-1 && Character.isLowerCase(aInput.charAt(i+1)))
				{	result.append(" ");
				}
			}
			result.append(c);
			lastUpper=upper;
			lastDigit=digit;
		}
		return result.toString();
	}

	@Override
	public void spaceChanged()
	{	mIndexBuilt.set(false);
		mIndexDir=null;
	}	
	
}
